# sage_setup: distribution = sagemath-categories
"""
Precision Error

The errors in this file indicate various styles of precision problems
that can go wrong for `p`-adics and power series.

AUTHORS:

- David Roe
"""

#*****************************************************************************
#       Copyright (C) 2008 David Roe <roed.math@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************


class PrecisionError(ArithmeticError):
    pass
