# sage_setup: distribution = sagemath-categories
from sage.misc.lazy_import import lazy_import
lazy_import('sage.sets.real_set', 'RealSet')
from sage.sets.set import Set
from sage.sets.integer_range import IntegerRange
from sage.sets.non_negative_integers import NonNegativeIntegers
from sage.sets.positive_integers import PositiveIntegers
from sage.sets.finite_enumerated_set import FiniteEnumeratedSet
lazy_import('sage.sets.recursively_enumerated_set', 'RecursivelyEnumeratedSet')
from sage.sets.totally_ordered_finite_set import TotallyOrderedFiniteSet
from sage.sets.disjoint_union_enumerated_sets import DisjointUnionEnumeratedSets
from sage.sets.primes import Primes
from sage.sets.family import Family
from sage.sets.disjoint_set import DisjointSet
from sage.sets.condition_set import ConditionSet
from sage.sets.finite_set_maps import FiniteSetMaps
del lazy_import
