# sage_setup: distribution = sagemath-categories
from sage.rings.function_field.all__sagemath_modules import *

try:
    from sage.rings.function_field.all__sagemath_symbolics import *
except ImportError:
    pass

from sage.misc.lazy_import import lazy_import

lazy_import("sage.rings.function_field.drinfeld_modules.drinfeld_module", "DrinfeldModule")
del lazy_import
